<?php
/**
 * @package     CSVI
 * @subpackage  HikaShop
 *
 * @author      RolandD Cyber Produksi <contact@rolandd.com>
 * @copyright   Copyright (C) 2006 - @year@ RolandD Cyber Produksi. All rights reserved.
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 * @link        https://rolandd.com
 */

use Joomla\CMS\Language\Text;

defined('_JEXEC') or die;

/**
 * The HikaShop category helper class.
 *
 * @package     CSVI
 * @subpackage  HikaShop
 * @since       6.0
 */
class Com_HikashopHelperCom_Hikashop_Category
{
	/**
	 * Database connector
	 *
	 * @var    JDatabaseDriver
	 * @since  6.0
	 */
	protected $db = null;

	/**
	 * CSVI template helper
	 *
	 * @var    CsviHelperTemplate
	 * @since  6.0
	 */
	protected $template = null;

	/**
	 * CSVI log helper
	 *
	 * @var    CsviHelperLog
	 * @since  6.0
	 */
	protected $log = null;

	/**
	 * The HikashopTableCategory table
	 *
	 * @var    HikashopTableCategory
	 * @since  6.0
	 */
	protected $table = null;

	/**
	 * Category separator
	 *
	 * @var    string
	 * @since  6.0
	 */
	protected $categorySeparator = null;

	/**
	 * Cache for processed category paths.
	 *
	 * @var    array
	 * @since  3.0
	 */
	protected $categoryCache = [];

	/**
	 * Array of categories to process
	 *
	 * @var    array
	 * @since  6.0
	 */
	private $categories = [];

	/**
	 * Constructor.
	 *
	 * @param   JDatabaseDriver $db Database connector.
	 *
	 * @throws  CsviException
	 *
	 * @since   4.0
	 */
	public function __construct($db)
	{
		$this->db = $db;

		// Load the table file
		require_once JPATH_PLUGINS . '/csviaddon/hikashop/com_hikashop/table/category.php';

		// Instantiate the table
		try
		{
			$this->table = new HikashopTableCategory('', '', $db);
		} catch (Exception $e)
		{
			throw new CsviException(Text::sprintf('COM_CSVI_NO_TABLE_LOADED', 'HikashopTableCategory', $e->getMessage()), 513);
		}
	}

	/**
	 * Save a category.
	 *
	 * A category element
	 * stdClass Object
	 * (
	 * [category_name] => Peripherals
	 * [category_meta_description] =>
	 * [category_keywords] =>
	 * [category_page_title] =>
	 * [category_alias] => peripherals
	 * [category_canonical] =>
	 * [category_published] => 1
	 * [category_layout] =>
	 * [category_quantity_layout] =>
	 * [category_parent_id] => 17
	 * [category_access] => all
	 * [category_id] => 18
	 * [category_type] =>
	 * [category_description] =>
	 * )
	 *
	 * @param   object $element  The array with information to save.
	 * @param   string $type     The type of category.
	 * @param   bool   $ordering Set if the categories should be re-ordered.
	 *
	 * @return  bool  True on success | False on failure.
	 *
	 * @since   6.0
	 */
	public function save($element, $type = 'product', $ordering = true)
	{
		$recalculate      = false;
		$new              = true;
		$newParentElement = new stdClass;

		// Clear the table
		$this->table->reset();

		if (!empty($element->category_id))
		{
			// Load the existing category
			$this->table->load($element->category_id);

			// Set the new element
			$new = false;

			// Load the old element
			$query = $this->db->getQuery(true)
				->select('*')
				->from($this->db->quoteName('#__hikashop_category'))
				->where($this->db->quoteName('category_id') . ' = ' . (int) $element->category_id);
			$this->db->setQuery($query, 0, 1);
			$old = $this->db->loadObject();

			if (isset($element->category_parent_id))
			{
				// Load the parent element
				$query = $this->db->getQuery(true)
					->select('*')
					->from($this->db->quoteName('#__hikashop_category'))
					->where($this->db->quoteName('category_id') . ' = ' . (int) $element->category_parent_id);
				$this->db->setQuery($query, 0, 1);
				$newParentElement = $this->db->loadObject();

				if ($old->category_parent_id != $element->category_parent_id)
				{
					if (($newParentElement->category_left > $old->category_left) && ($newParentElement->category_right < $old->category_right))
					{
						return false;
					}

					$recalculate = true;
				}

				if (!empty($newParentElement->category_type) && $newParentElement->category_type != 'root')
				{
					$element->category_type = $newParentElement->category_type;
				}

				if (!empty($element->category_site_id) && $newParentElement->category_type == 'root')
				{
					$element->category_site_id = '';
				}
			}

			if (empty($element->category_type))
			{
				if (!empty($old->category_type))
				{
					if ($old->category_type == 'root')
					{
						$element->category_type = $type;
					}
					else
					{
						$element->category_type = $old->category_type;
					}
				}
				else
				{
					$element->category_type = $type;
				}
			}

			$element->category_modified = time();
		}
		else
		{
			if (empty($element->category_parent_id))
			{
				$query = $this->db->getQuery(true)
					->select($this->db->quoteName('category_id'))
					->from($this->db->quoteName('#__hikashop_category'))
					->where($this->db->quoteName('category_type') . ' = ' . $this->db->quote('root'));
				$this->db->setQuery($query, 0, 1);
				$element->category_parent_id = $this->db->loadResult();
				$element->category_created   = time();
				$element->category_namekey   = $element->category_type . '_' . $element->category_created . '_' . rand();
				$element->category_depth     = 1;
			}

			// Load the new element
			$query = $this->db->getQuery(true)
				->select('*')
				->from($this->db->quoteName('#__hikashop_category'))
				->where($this->db->quoteName('category_id') . ' = ' . (int) $element->category_parent_id);
			$this->db->setQuery($query, 0, 1);
			$newParentElement = $this->db->loadObject();

			if (empty($element->category_type) && $newParentElement->category_type != 'root')
			{
				$element->category_type = $newParentElement->category_type;
			}

			if (empty($element->category_type))
			{
				$element->category_type = $type;
			}

			if (empty($element->category_site_id) && $newParentElement->category_type != 'root')
			{
				$element->category_site_id = $newParentElement->category_site_id;
			}

			$element->category_created = $element->category_modified = time();

			if (empty($element->category_namekey))
			{
				$element->category_namekey = $newParentElement->category_type . '_' . $element->category_created . '_' . rand();
			}

			if (!isset($element->category_published))
			{
				$element->category_published = 1;
			}

			if (!isset($element->category_description))
			{
				$element->category_description = '';
			}

			if (!isset($element->category_keywords))
			{
				$element->category_keywords = '';
			}

			if (!isset($element->category_meta_description))
			{
				$element->category_meta_description = '';
			}
		}

		// Save the category
		if (!$this->table->save($element))
		{
			return false;
		}

		if (empty($element->category_id))
		{
			$element->category_id = $this->table->get('category_id');

			if ($ordering)
			{
				$orderClass                     = new stdClass;
				$orderClass->pkey               = 'category_id';
				$orderClass->table              = 'category';
				$orderClass->groupMap           = 'category_parent_id';
				$orderClass->category_parent_id = $element->category_parent_id;
				$orderClass->orderingMap        = 'category_ordering';
				$this->reOrder($orderClass);
			}
		}

		if ($new)
		{
			$query = $this->db->getQuery(true)
				->update($this->db->quoteName('#__hikashop_category'))
				->set($this->db->quoteName('category_right') . ' = ' . $this->db->quoteName('category_right') . ' + 2')
				->where($this->db->quoteName('category_right') . ' >= ' . (int) $newParentElement->category_right);
			$this->db->setQuery($query)->execute();

			$query->clear()
				->update($this->db->quoteName('#__hikashop_category'))
				->set($this->db->quoteName('category_left') . ' = ' . $this->db->quoteName('category_left') . ' + 2')
				->where($this->db->quoteName('category_left') . ' >= ' . (int) $newParentElement->category_right);
			$this->db->setQuery($query)->execute();

			$query->clear()
				->update($this->db->quoteName('#__hikashop_category'))
				->set($this->db->quoteName('category_left') . ' = ' . (int) $newParentElement->category_right)
				->set($this->db->quoteName('category_right') . ' = ' . (int) ($newParentElement->category_right + 1))
				->set($this->db->quoteName('category_depth') . ' = ' . (int) ($newParentElement->category_depth + 1))
				->where($this->db->quoteName('category_id') . ' = ' . $this->table->get('category_id'));
			$this->db->setQuery($query)->execute();
		}
		elseif ($recalculate)
		{
			$query = $this->db->getQuery(true)
				->select(
					array(
						$this->db->quoteName('category_left'),
						$this->db->quoteName('category_right'),
						$this->db->quoteName('category_depth'),
						$this->db->quoteName('category_id'),
						$this->db->quoteName('category_parent_id')

					)
				)
				->from($this->db->quoteName('#__hikashop_category'))
				->order($this->db->quoteName('category_left') . ' ASC');
			$this->db->setQuery($query);
			$categories       = $this->db->loadObjectList();
			$root             = null;
			$this->categories = array();

			foreach ($categories as $cat)
			{
				$this->categories[$cat->category_parent_id][] = $cat;

				if (empty($cat->category_parent_id))
				{
					$root = $cat;
				}
			}

			$this->rebuildTree($root, 0, 1);
		}

		return $this->table->get('category_id');
	}

	/**
	 * Find the root ID.
	 *
	 * @return  int  The ID of the root.
	 *
	 * @since   6.0
	 */
	protected function getRoot()
	{
		// Make the $id static since it won't change
		static $id = 0;

		if (empty($id))
		{
			$query = $this->db->getQuery(true)
				->select($this->db->quoteName('category_id'))
				->from($this->db->quoteName('#__hikashop_category'))
				->where($this->db->quoteName('category_type') . ' = ' . $this->db->quote('root'));
			$this->db->setQuery($query, 0, 1);
			$id = $this->db->loadResult();
		}

		return $id;
	}

	/**
	 * Find the ID of a main element.
	 *
	 * @param   string &$element The type to get the root ID for.
	 *
	 * @return  int  The ID of the main element.
	 *
	 * @since   6.0
	 */
	private function getMainElement(&$element)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('category_id'))
			->from($this->db->quoteName('#__hikashop_category'))
			->where($this->db->quoteName('category_parent_id') . ' = ' . (int) $this->getRoot())
			->where($this->db->quoteName('category_type') . ' = ' . $this->db->quote($element));
		$this->db->setQuery($query);

		$element = (array) $this->db->loadObjectList();
	}

	/**
	 * Rebuild category tree.
	 *
	 * @param   object $element The element to rebuild.
	 * @param   int    $depth   The depth of the element.
	 * @param   int    $left    The position of the element.
	 *
	 * @return  array  The depth and left value.
	 *
	 * @since   6.0
	 */
	private function rebuildTree($element, $depth, $left)
	{
		$currentLeft  = $left;
		$currentDepth = $depth;

		if (!empty($this->categories[$element->category_id]))
		{
			$depth++;

			foreach ($this->categories[$element->category_id] as $child)
			{
				$left++;
				list($depth, $left) = $this->rebuildTree($child, $depth, $left);
			}

			$depth--;
		}

		$left++;

		if ($currentLeft != $element->category_right || $currentLeft != $element->category_left || $currentDepth != $element->category_depth)
		{
			$query = $this->db->getQuery(true)
				->update($this->db->quoteName('#__hikashop_category'))
				->set($this->db->quoteName('category_left') . ' = ' . (int) $currentLeft)
				->set($this->db->quoteName('category_right') . ' = ' . (int) $left)
				->set($this->db->quoteName('category_depth') . ' = ' . (int) $currentDepth)
				->where($this->db->quoteName('category_id') . ' = ' . (int) $element->category_id);
			$this->db->setQuery($query)->execute();
		}

		return array($depth, $left);
	}

	/**
	 * Reorder the categories.
	 *
	 * @param   object $orderObject The reorder details.
	 *
	 * @return  void.
	 *
	 * @since   6.0
	 */
	private function reOrder($orderObject)
	{
		$orderingMap = $orderObject->orderingMap;

		$query = $this->db->getQuery(true)
			->select('MAX(' . $this->db->quoteName($orderingMap) . ')')
			->from($this->db->quoteName('#__hikashop_category'))
			->where($this->db->quoteName('category_parent_id') . ' = ' . $orderObject->category_parent_id);
		$this->db->setQuery($query);
		$max = $this->db->loadResult();
		$max++;

		$query->clear()
			->update($this->db->quoteName('#__hikashop_category'))
			->set($this->db->quoteName($orderingMap) . ' = ' . (int) $max)
			->where($this->db->quoteName($orderingMap) . ' = 0')
			->where($this->db->quoteName('category_parent_id') . ' = ' . $orderObject->category_parent_id);
		$this->db->setQuery($query)->execute();

		$query->clear()
			->select($this->db->quoteName($orderingMap) . ',' . $this->db->quoteName('category_id'))
			->from($this->db->quoteName('#__hikashop_category'))
			->where($this->db->quoteName('category_parent_id') . ' = ' . $orderObject->category_parent_id)
			->order($this->db->quoteName($orderingMap) . ' ASC');
		$this->db->setQuery($query);
		$results = $this->db->loadObjectList();
		$i       = 1;

		if (!empty($results))
		{
			foreach ($results as $oneResult)
			{
				if ($oneResult->$orderingMap != $i)
				{
					$oneResult->$orderingMap = $i;
					$this->db->updateObject('#__hikashop_category', $oneResult, 'category_id');
				}

				$i++;
			}
		}
	}

	/**
	 * Retrieve category ID based on path, creating them if not exists.
	 *
	 * @param   string $path  The path to get the ID for.
	 * @param   string $group The group the category belongs to.
	 *
	 * @return  int  The category ID.
	 *
	 * @since   6.0
	 */
	public function getCategoryIdByPath($path, $group = 'product')
	{
		// Explode the path
		$category_list  = explode($this->categorySeparator, $path);
		$category_count = count($category_list);

		// Initiate the category ID
		$category_id = 0;

		// Get the root ID
		$category_parent_id = $this->getRoot();

		// Loop through the categories to find the ID for the last category
		for ($i = 0; $i < $category_count; $i++)
		{
			// Check the cache first
			if (array_key_exists($category_parent_id . '.' . $category_list[$i], $this->categoryCache))
			{
				$category_id = $this->categoryCache[$category_parent_id . '.' . $category_list[$i]];
			}
			else
			{
				if ($category_list[$i])
				{
					// See if this category exists with it's parent in xref
					$query = $this->db->getQuery(true)
						->select($this->db->quoteName('c.category_id'))
						->from($this->db->quoteName('#__hikashop_category', 'c'))
						->leftJoin(
							$this->db->quoteName('#__hikashop_category', 'p')
							. ' ON ' . $this->db->quoteName('c.category_id') . ' = ' . $this->db->quoteName('p.category_id')
						)
						->where(
							$this->db->quoteName('c.category_left')
							. ' BETWEEN '
							. $this->db->quoteName('p.category_left')
							. ' AND '
							. $this->db->quoteName('p.category_right')
						)
						->where($this->db->quoteName('c.category_name') . ' = ' . $this->db->quote($category_list[$i]) . ' OR ' .
							$this->db->quoteName('c.category_alias') . ' = ' . $this->db->quote($category_list[$i]))
						->where($this->db->quoteName('c.category_parent_id') . ' = ' . (int) $category_parent_id);
					$this->db->setQuery($query);
					$category_id = $this->db->loadResult();

					$this->log->add('Check category exists: ' . $category_list[$i]);
				}

				// Add result to cache if category exists
				if ($category_id)
				{
					$this->categoryCache[$category_parent_id . '.' . $category_list[$i]] = $category_id;
				}
			}

			// Category does not exist, create it
			if (is_null($category_id) && $category_list[$i] && !$this->template->get('ignore_non_exist', false))
			{
				$category = (object) array(
					'category_id'        => null,
					'category_parent_id' => $category_parent_id,
					'category_name'      => $category_list[$i],
					'category_type'      => $group,
					'category_created'   => time(),
					'category_modified'  => time(),
					'category_published' => 1
				);

				$category_id = $this->save($category);
			}

			// Set this category as parent of next in line
			$category_parent_id = $category_id;
		}

		return $category_id;
	}

	/**
	 * Create a category path based on the category ID.
	 *
	 * @param   mixed $catids An array or int of category IDs to create paths for.
	 *
	 * @return  string  A pipe separated list of category paths.
	 *
	 * @since   6.0
	 */
	public function createCategoryPathById($catids)
	{
		$paths = '';

		if (!is_array($catids))
		{
			$catids = array($catids);
		}

		if (!empty($catids))
		{
			$catpaths = $this->constructCategoryPath($catids);
			$paths    = implode($this->template->get('items_separator', '|'), $catpaths);
		}

		return $paths;
	}

	/**
	 * Create a category path or array of category paths based on the product ID.
	 *
	 * @param   int  $product_id The ID to create the paths for.
	 * @param   bool $id         Set if the category IDs should be returned.
	 *
	 * @return  string  A pipe separated list of category paths or category IDs.
	 *
	 * @since   6.0
	 *
	 * @throws  RuntimeException
	 */
	public function createCategoryPath($product_id, $id = false)
	{
		$paths = '';

		// Get the category paths
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('category_id'))
			->from($this->db->quoteName('#__hikashop_product_category'))
			->where($this->db->quoteName('product_id') . ' = ' . (int) $product_id);
		$this->db->setQuery($query);
		$catids = $this->db->loadColumn();

		if (!empty($catids))
		{
			// Return the paths
			if ($id)
			{
				$paths = implode($this->template->get('items_separator', '|'), $catids);
			}
			else
			{
				$catpaths = $this->constructCategoryPath($catids);

				$paths = implode($this->template->get('items_separator', '|'), $catpaths);
			}
		}

		return $paths;
	}

	/**
	 * Construct the category path.
	 *
	 * @param   array $catids The IDs to build a category for.
	 *
	 * @return  array  An array of category paths.
	 *
	 * @since   4.0
	 */
	public function constructCategoryPath($catids)
	{
		$catpaths = array();

		if (is_array($catids))
		{
			// Get the paths
			foreach ($catids as $category_id)
			{
				/**
				 * Create the path
				 * SELECT category_left,category_right FROM jos_hikashop_category WHERE category_id IN (19);
				 * SELECT category_name from jos_hikashop_category where category_left <= 16 and category_right >= 17;
				 */

				// Get the leafs
				$leafs = $this->getLeafs($category_id);

				if (!empty($leafs))
				{
					if (!isset($this->categoryCache[$leafs->category_left . '.' . $leafs->category_right]))
					{
						// Load the path
						$query = $this->db->getQuery(true)
							->select($this->db->quoteName('category_name'))
							->from($this->db->quoteName('#__hikashop_category'))
							->where($this->db->quoteName('category_left') . ' <= ' . (int) $leafs->category_left)
							->where($this->db->quoteName('category_left') . ' > 1 ')
							->where($this->db->quoteName('category_right') . ' >= ' . (int) $leafs->category_right)
							->order($this->db->quoteName('category_left'));
						$this->db->setQuery($query);
						$paths = $this->db->loadColumn();

						// Create the path
						$catpaths[] = implode($this->categorySeparator, $paths);

						// Add to the cache
						$this->categoryCache[$leafs->category_left . '.' . $leafs->category_right] = implode($this->categorySeparator, $paths);
					}
					else
					{
						$catpaths[] = $this->categoryCache[$leafs->category_left . '.' . $leafs->category_right];
					}
				}
			}
		}

		return $catpaths;
	}

	/**
	 * Get the children for an element.
	 *
	 * @param   int    $element           Set the ID for which the children should be returned.
	 * @param   bool   $all               Set if all children should be returned.
	 * @param   array  $additionalFilters Additional filters to apply.
	 * @param   string $order             Set the field to order by.
	 * @param   int    $start             Set the start position to get values from.
	 * @param   int    $value             Set the number of items to return.
	 * @param   bool   $category_image    Should the category image be returned as well.
	 * @param   string $select            Set which fields to select.
	 * @param   bool   $orderTableJoin    Set if the order table should be joined.
	 *
	 * @return  array  List of children.
	 *
	 * @since   6.0
	 */
	public function getChildren(
		$element,
		$all = false,
		$additionalFilters = [],
		$order = '',
		$start = 0,
		$value = 20,
		$category_image = false,
		$select = 'a.*',
		$orderTableJoin = false)
	{
		$filters     = [];
		$treeFilters = [];

		if (empty($element))
		{
			$element = $this->getRoot();
		}

		if (is_array($element))
		{
			if (count($element) > 1)
			{
				if ($all)
				{
					$query = $this->db->getQuery(true)
						->select($this->db->quoteName(array('category_left', 'category_right')))
						->from($this->db->quoteName('#__hikashop_category'))
						->where($this->db->quoteName('category_id') . ' IN (' . implode(',', $element) . ')');
					$this->db->setQuery($query);
					$leafs      = $this->db->loadObjectList();
					$conditions = array();

					foreach ($leafs as $v)
					{
						$conditions[] = '(a.category_left > ' . $v->category_left . ' AND a.category_right < ' . $v->category_right . ')';
					}

					$filters[] = '(' . implode(' OR ', $conditions) . ')';
					$type      = 0;
				}
				else
				{
					foreach ($element as $k => $v)
					{
						$element[$k] = (int) $v;
					}

					$filters[] = 'a.category_parent_id IN (' . implode(',', $element) . ')';
					$type      = 0;
				}
			}
			else
			{
				$element = (int) array_pop($element);
			}
		}
		elseif (!is_numeric($element))
		{
			$this->getMainElement($element);
		}

		if (is_array($element))
		{
			if ($all)
			{
				foreach ($element as $eachCategory)
				{
					$query = $this->db->getQuery(true)
						->select('*')
						->from($this->db->quoteName('#__hikashop_category'))
						->where($this->db->quoteName('category_id') . ' = ' . (int) $eachCategory->category_id);
					$this->db->setQuery($query, 0, 1);
					$data = $this->db->loadObject();

					if (!empty($data))
					{
						if ((int) ($data->category_left + 1) === (int) $data->category_right)
						{
							continue;
						}

						$treeFilters[] = 'a.category_left >= ' . $data->category_left . ' AND a.category_right <=' . $data->category_right;

						if (!empty($data->category_type) && $data->category_type != 'root')
						{
							$type = $data->category_type;
						}
					}
				}

				if ($treeFilters)
				{
					$filters[] = '(' . implode(' OR ', $treeFilters) . ')';
				}
			}
			else
			{
				if ($element)
				{
					$filters[] = 'a.category_parent_id = ' . $element[0]->category_id;
					$type      = 0;
				}
			}
		}
		elseif (!is_array($element))
		{
			$type = $element;
		}

		if (!empty($type))
		{
			if ($type == 'product')
			{
				$filters[] = '(a.category_type = ' . $this->db->quote($type) . ' OR a.category_type = ' . $this->db->quote('vendor') . ')';
			}
			else
			{
				$filters[] = 'a.category_type = ' . $this->db->quote($type);
			}
		}

		if (!empty($additionalFilters))
		{
			$filters = array_merge($filters, $additionalFilters);
		}

		$query = $this->db->getQuery(true)
			->select($select)
			->from($this->db->quoteName('#__hikashop_category', 'a'));

		if (!empty($filters))
		{
			$query->where(implode(' AND ', $filters));
		}

		// Include order table for manufacturer and category filter for order export
		if ($orderTableJoin)
		{
			// For $element as 'order' element value will be 0, so check if it is order or manufacturer
			if ($element)
			{
				$query->leftJoin(
					$this->db->quoteName('#__hikashop_product', 'p')
					. ' ON ' . $this->db->quoteName('p.product_manufacturer_id') . ' = ' . $this->db->quoteName('a.category_id')
				);
				$query->leftJoin(
					$this->db->quoteName('#__hikashop_order_product', 'op')
					. ' ON ' . $this->db->quoteName('op.product_id') . ' = ' . $this->db->quoteName('p.product_id')
				);
				$query->group($this->db->quoteName('a.category_id'));
			}
			else
			{
				$query->leftJoin(
					$this->db->quoteName('#__hikashop_product_category', 'pc')
					. ' ON ' . $this->db->quoteName('pc.category_id') . ' = ' . $this->db->quoteName('a.category_id')
				);
				$query->rightJoin(
					$this->db->quoteName('#__hikashop_order_product', 'op')
					. ' ON ' . $this->db->quoteName('op.product_id') . ' = ' . $this->db->quoteName('pc.product_id')
				);

				$query->group($this->db->quoteName('a.category_id'));

				// Load the category for variants as well
				$variantCategoryQuery = $query;
				$variantCategoryQuery->clear('join');

				$variantCategoryQuery->leftJoin(
					$this->db->quoteName('#__hikashop_product_category', 'pc')
					. ' ON ' . $this->db->quoteName('pc.category_id') . ' = ' . $this->db->quoteName('a.category_id')
				);
				$variantCategoryQuery->leftJoin(
					$this->db->quoteName('#__hikashop_product', 'p')
					. ' ON ' . $this->db->quoteName('p.product_parent_id') . ' = ' . $this->db->quoteName('pc.product_id')
					. ' OR ' . $this->db->quoteName('p.product_id') . ' = ' . $this->db->quoteName('pc.product_id')
				);
				$variantCategoryQuery->rightJoin(
					$this->db->quoteName('#__hikashop_order_product', 'op')
					. ' ON ' . $this->db->quoteName('op.product_id') . ' = ' . $this->db->quoteName('p.product_id')
				);

				$query->union($variantCategoryQuery->__toString());
			}
		}

		if ($order)
		{
			$query->order($this->db->quoteName($order));
		}

		$this->db->setQuery($query, (int) $start, (int) $value);

		$rows = $this->db->loadObjectList();

		if ($category_image && !empty($rows))
		{
			$ids = array();

			foreach ($rows as $row)
			{
				$ids[] = $row->category_id;
			}

			$query->clear()
				->select('*')
				->from($this->db->quoteName('#__hikashop_file'))
				->where($this->db->quoteName('file_type') . ' = ' . $this->db->quote('category'))
				->where($this->db->quoteName('file_ref_id') . ' IN (' . implode(',', $ids) . ')');
			$this->db->setQuery($query);
			$images = $this->db->loadObjectList();

			foreach ($rows as $k => $cat)
			{
				if (!empty($images))
				{
					foreach ($images as $img)
					{
						if ($img->file_ref_id == $cat->category_id)
						{
							foreach (get_object_vars($img) as $key => $val)
							{
								$rows[$k]->$key = $val;
							}

							break;
						}
					}
				}

				if (!isset($rows[$k]->file_name))
				{
					if (isset($row))
					{
						$rows[$k]->file_name = $row->category_name;
					}
				}
			}
		}

		return $rows;
	}

	/**
	 * Get the leafs for a given category.
	 *
	 * @param   int $category_id The ID of the category to get the leafs for.
	 *
	 * @return  object  The leafs for a given category.
	 *
	 * @since   6.0
	 */
	public function getLeafs($category_id)
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName(array('category_left', 'category_right')))
			->from($this->db->quoteName('#__hikashop_category'))
			->where($this->db->quoteName('category_id') . ' = ' . (int) $category_id);
		$this->db->setQuery($query);

		return $this->db->loadObject();
	}

	/**
	 * Get the children for an element.
	 *
	 * @param   int $category_ids Set the ID for which the children should be returned.
	 *
	 * @return  array  List of children.
	 *
	 * @since   6.0
	 */
	public function getChildrenIds($category_ids)
	{
		if (!is_array($category_ids))
		{
			$category_ids = array($category_ids);
		}

		$query = $this->db->getQuery(true)
			->select($this->db->quoteName(array('category_left', 'category_right')))
			->from($this->db->quoteName('#__hikashop_category'))
			->where($this->db->quoteName('category_id') . ' IN (' . implode(',', $category_ids) . ')');
		$this->db->setQuery($query);
		$leafs      = $this->db->loadObjectList();
		$conditions = array();

		foreach ($leafs as $v)
		{
			$conditions[] = '(c.category_left > ' . $v->category_left . ' AND c.category_right < ' . $v->category_right . ')';
		}

		$filters[] = '(' . implode(' OR ', $conditions) . ')';

		return $this->db->getQuery(true)
			->select($this->db->quoteName('c.category_id'))
			->from($this->db->quoteName('#__hikashop_category', 'c'))
			->where($filters);
	}

	/**
	 * Delete a given category.
	 *
	 * @param   int    $category_id   The ID of the category to delete.
	 * @param   string $category_type The type of category.
	 *
	 * @return  void.
	 *
	 * @since   6.0
	 */
	public function delete($category_id, $category_type)
	{
		// Get the position of the category
		$query = $this->db->getQuery(true)
			->select(
				$this->db->quoteName(
					[
						'category_left',
						'category_right',
					]
				)
			)
			->select('(' . $this->db->quoteName('category_right') . ' - ' . $this->db->quoteName('category_left') . ') AS ' . $this->db->quoteName('width'))
			->from($this->db->quoteName('#__hikashop_category'))
			->where($this->db->quoteName('category_id') . ' = ' . (int) $category_id);
		$this->db->setQuery($query);
		$data = $this->db->loadObject();

		// Delete the category
		$query->clear()
			->delete($this->db->quoteName('#__hikashop_category'))
			->where($this->db->quoteName('category_left') . ' BETWEEN ' . (int) $data->category_left . ' AND ' . (int) $data->category_right);
		$this->db->setQuery($query)->execute();

		// Update the right nodes
		$query->clear()
			->update($this->db->quoteName('#__hikashop_category'))
			->set($this->db->quoteName('category_right') . ' = ' . $this->db->quoteName('category_right') . ' - ' . (int) $data->width)
			->where($this->db->quoteName('category_right') . ' > ' . $data->category_right);
		$this->db->setQuery($query)->execute();

		// Update the left nodes
		$query->clear()
			->update($this->db->quoteName('#__hikashop_category'))
			->set($this->db->quoteName('category_left') . ' = ' . $this->db->quoteName('category_left') . ' - ' . (int) $data->width)
			->where($this->db->quoteName('category_left') . ' > ' . $data->category_right);
		$this->db->setQuery($query)->execute();
	}
}
